"""Toolbox providing user interface for the tools."""

import os

import arcpy
import nast
import csa


class Toolbox(object):
    """Required class containing all the tools in the toolbox"""

    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Toolbox"
        self.alias = ""

        # List of tool classes associated with this toolbox
        self.tools = [ComputeServiceAreas]


class ComputeServiceAreas(object):
    """User interface for the tool"""

    NETWORK_SOURCE_PARAM_INDEX = 3
    TRAVEL_MODE_PARAM_INDEX = 4
    SERVICE_AREAS_PARAM_INDEX = 5

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Compute Service Areas"
        self.description = "Compute service areas using ArcGIS Online Service Areas service or a local network dataset."
        self.canRunInBackground = False  # pylint: disable=invalid-name

    def getParameterInfo(self):  # pylint: disable=invalid-name
        """Define parameter definitions"""

        # Facilities parameter
        facilities_param = arcpy.Parameter("facilities", "Facilities", "Input", "GPFeatureLayer", "Required")
        facilities_param.filter.list = ["Point"]

        # Break Values parameter
        break_values_param = arcpy.Parameter("break_values", "Break Values", "Input", "GPDouble", "Required",
                                             multiValue=True)
        break_values_param.value = [5, 10, 15]

        # Break Units parameter
        break_units_param = arcpy.Parameter("break_units", "Break Units", "Input", "GPString", "Required")
        break_units_param.filter.list = ["Seconds", "Minutes", "Hours", "Feet", "Meters", "Kilometers", "Miles"]
        break_units_param.value = "Minutes"

        # Network Source parameter
        network_data_source_parameter = arcpy.Parameter("network_data_source", "Network Data Source", "Input",
                                                        ["GPNetworkDatasetLayer", "GPString"], "Required")
        network_data_source_parameter.value = "ARCGIS_ONLINE"

        # Travel mode parameter
        travel_mode_parameter = arcpy.Parameter("travel_mode", "Travel Mode", "Input", "GPString", "Required")
        travel_mode_parameter.filter.list = []

        # Output Service Areas parameter
        output_service_areas_param = arcpy.Parameter("output_service_areas", "Output Service Areas", "Output",
                                                     "DEFeatureClass", "Required")
        output_service_areas_param.value = "ServiceAreas"
        output_service_areas_param.symbology = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                                            "data", "ServiceAreas.lyr")

        params = [facilities_param, break_values_param, break_units_param, network_data_source_parameter,
                  travel_mode_parameter, output_service_areas_param]

        return params

    def isLicensed(self):  # pylint: disable=invalid-name
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):  # pylint: disable=invalid-name
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""

        # Populate list of travel modes and the default travel mode value  based on the network source
        travel_mode_param = parameters[self.TRAVEL_MODE_PARAM_INDEX]
        network_source_param = parameters[self.NETWORK_SOURCE_PARAM_INDEX]
        network_source = network_source_param.value

        if network_source:
            try:
                travel_modes = nast.NATool.get_travel_modes(network_source)
            except (ValueError, arcpy.ExecuteError):
                travel_mode_param.filter.list = []
                travel_mode_param.value = ""
                return
            supported_travel_modes = travel_modes["supportedTravelModes"]
            default_travel_mode_name = travel_modes["defaultTravelModeName"]
            if supported_travel_modes:
                travel_mode_param.filter.list = sorted(list(supported_travel_modes))
            else:
                travel_mode_param.filter.list = []
                travel_mode_param.value = ""
            # Set a default value for the travel mode parameter if a value has not been set
            if not travel_mode_param.value:
                travel_mode_param.value = default_travel_mode_name
            # Set a default value for the travel mode parameter if a value is not selected by the user.
            if not travel_mode_param.altered:
                travel_mode_param.value = default_travel_mode_name

        return

    def updateMessages(self, parameters):  # pylint: disable=invalid-name, unused-argument
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""

        network_source_param = parameters[self.NETWORK_SOURCE_PARAM_INDEX]
        travel_mode_param = parameters[self.TRAVEL_MODE_PARAM_INDEX]

        # Check if signed in to ArcGIS Online or a local network source is valid and supports travel modes.
        nast.NATool.verify_network_source(network_source_param, travel_mode_param)

        # Check if travel mode is a valid travel mode JSON or one of the supported names
        nast.NATool.verify_travel_mode(travel_mode_param)

        return

    def execute(self, parameters, messages):  # pylint: disable=invalid-name, unused-argument
        """The source code of the tool."""
        # Read the parameter values
        facilities = parameters[0].valueAsText
        break_values = parameters[1].values
        break_units = parameters[2].valueAsText
        network_source = parameters[self.NETWORK_SOURCE_PARAM_INDEX].value
        travel_mode = parameters[self.TRAVEL_MODE_PARAM_INDEX].valueAsText
        service_areas = parameters[self.SERVICE_AREAS_PARAM_INDEX].valueAsText

        # Get the token for the signed in user.
        signin_token = arcpy.GetSigninToken()
        # Pass the parameters to the core class.
        tool = csa.ComputeServiceAreas(facilities, break_values, travel_mode, service_areas, signin_token, break_units,
                                       network_source)
        # Execute the core class.
        try:
            tool.execute()
            # Set the value for the tool output parameter.
            arcpy.SetParameterAsText(self.SERVICE_AREAS_PARAM_INDEX, tool.service_areas)
        except arcpy.ExecuteError as ex:
            arcpy.AddError("A geoprocessing error occurred during tool execution.")
            arcpy.AddError(str(ex))

        return
