"""Contains core class that performs execution logic for ComputeServiceAreas tool."""

import locale
import time

import arcpy
import nast


class ComputeServiceAreas(nast.NATool):
    """Computes service areas around facilities based on local or ArcGIS Online network source."""

    def __init__(self, facilities, break_values, travel_mode, service_areas, sign_in_token, break_units="Minutes",
                 network_source="ARCGIS_ONLINE"):
        """Store tool parameter values and define names that are used in other methods."""
        self.facilities = facilities
        self.break_values = break_values
        self.break_units = break_units
        self.network_source = network_source
        self.travel_mode = travel_mode
        self.service_areas = service_areas
        self.sign_in_token = sign_in_token
        self.breaks = ""

    def execute(self):
        """Provide the logic for computing service areas."""
        # Convert break units from a list of doubles to a space separated string
        self.breaks = " ".join([locale.str(val) for val in self.break_values])
        if self.network_source == "ARCGIS_ONLINE":
            self._execute_portal_network_source()
        else:
            self._execute_local_network_source()

    def _execute_local_network_source(self):
        """Compute service areas using a local network source.

        The method raises arcpy.ExecuteError exception when an error occurs during execution.
        """
        service_area_layer_name = "ServiceAreas"
        search_tolerance = "20000 Meters"
        travel_mode = None

        # Check if we have ArcGIS Desktop Basic and Network Extension license
        try:
            nast.NATool.check_license("arcview", "network")
            arcpy.CheckOutExtension("network")
        except nast.LicenseError as ex:
            arcpy.AddIDMessage("ERROR", 626, "Compute Service Areas")
            arcpy.AddError(str(ex))
            raise arcpy.ExecuteError

        # Get travel mode object corresonding to the travel mode name
        try:
            travel_modes = nast.NATool.get_travel_modes(self.network_source)
        except ValueError as ex:
            raise arcpy.ExecuteError(u"{}\n{}".format(nast.NATool.MSG_INVALID_NETWORK_SOURCE, ex))
        nds_travel_modes = travel_modes["supportedTravelModes"]
        # Fail if the network dataset does not support travel modes
        if not nds_travel_modes:
            raise arcpy.ExecuteError(nast.NATool.MSG_INVALID_NETWORK_SOURCE)
        if self.travel_mode in nds_travel_modes:
            travel_mode = nds_travel_modes[self.travel_mode]
        else:
            # Fail if the travel mode JSON string can not be converted into a travel mode object
            try:
                travel_mode = arcpy.na.TravelMode(self.travel_mode)
            except ValueError:
                arcpy.AddIDMessage("ERROR", 30158, self.travel_mode, u" | ".join(list(nds_travel_modes)))
                raise arcpy.ExecuteError
        # Get the network dataset impedance from the travel mode object
        impedance_attribute = travel_mode.impedance

        # Create a new service area layer
        arcpy.AddMessage("Creating service area analysis layer")
        result = arcpy.na.MakeServiceAreaLayer(self.network_source, service_area_layer_name, impedance_attribute,
                                               "TRAVEL_FROM", self.breaks)
        # Get the layer object from the result object. The service area layer can now be referenced using the layer
        # object.
        service_area_layer = result.getOutput(0)  # pylint: disable=no-member

        # Get the names of all the sublayers within the service area layer.
        sub_layer_names = arcpy.na.GetNAClassNames(service_area_layer)

        # Stores the layer names that we will use later
        facilities_layer_name = sub_layer_names["Facilities"]
        polygons_layer_name = sub_layer_names["SAPolygons"]

        # Get the layer object for the output polygons.
        if nast.NATool.is_arcgis_pro():
            # When using ArcGIS Pro, get the polygons sub layer using the listLayers method on the service area layer
            # object
            polygons_layer = service_area_layer.listLayers(polygons_layer_name)[0]
        else:
            # When using ArcMap, get the polygons sub layer using the ListLayers function in the mapping module
            polygons_layer = arcpy.mapping.ListLayers(service_area_layer, polygons_layer_name)[0]

        # Map the StoreName field from stores as Name field using field mappings
        field_mappings = arcpy.na.NAClassFieldMappings(service_area_layer, facilities_layer_name)
        field_mappings["Name"].mappedFieldName = "StoreName"
        # Load the stores as facilities using the field mappings.
        arcpy.AddMessage("Loading facilities")
        arcpy.na.AddLocations(service_area_layer, facilities_layer_name, self.facilities, field_mappings,
                              search_tolerance)

        # Update the analysis settings for the service area layer based on the travel mode.
        service_area_layer_properties = arcpy.na.GetSolverProperties(service_area_layer)
        service_area_layer_properties.applyTravelMode(travel_mode)
        # Enable trimming of the output polygons
        service_area_layer_properties.trimPolygons = "TRIM_POLYS"
        service_area_layer_properties.trimDistance = "100 Meters"

        # Solve the service area layer. Fail if the solve returned a failure.
        try:
            arcpy.AddMessage("Performing service area analysis")
            arcpy.na.Solve(service_area_layer)
        except arcpy.ExecuteError:
            err_msgs = arcpy.GetMessages(2)
            raise arcpy.ExecuteError(u"An error occurred when running the solve tool.\n{}".format(err_msgs))

        # Save the service areas to the output feature class.
        arcpy.AddMessage("Saving the results.")
        arcpy.management.CopyFeatures(polygons_layer, self.service_areas)

    def _execute_portal_network_source(self):
        """Compute service areas using ArcGIS Online network source.

        The method raises arcpy.ExecuteError exception when an error occurs during execution.
        """
        # Fail if the token of a signed in user is not specified
        if not self.sign_in_token:
            arcpy.AddError(nast.NATool.MSG_NOT_SIGNED_IN)
            raise arcpy.ExecuteError

        # Import the ArcGIS Online Service Areas service as a tool
        arcpy.AddMessage("Importing ArcGIS Online service as a tool.")
        service = nast.NATool.import_service("World/ServiceAreas", token=self.sign_in_token["token"],
                                             referer=self.sign_in_token["referer"])

        # Call the service
        arcpy.AddMessage("Executing the service tool.")
        result = service.GenerateServiceAreas(self.facilities, self.breaks, self.break_units,
                                              Travel_Mode=self.travel_mode)
        # Check the status of the result object every second until it has a value of 4(succeeded) or greater
        while result.status < 4:
            time.sleep(1)

        # Print any warning or error messages returned from the tool
        result_severity = result.maxSeverity
        if result_severity == 2:
            err_msgs = result.getMessages(2)
            raise arcpy.ExecuteError(u"An error occurred when running the service tool.\n{}".format(err_msgs))
        elif result_severity == 1:
            arcpy.AddMessage("Warnings were returned when running the service tool.")
            arcpy.AddWarning(result.getMessages(1))

        # Save the service areas to the output feature class.
        arcpy.AddMessage("Saving the results from the service tool.")
        result.getOutput(0).save(self.service_areas)
