"""Module containing common classes and functions for all the tools in NetworkAnalysisSampleTools toolbox."""

import hashlib
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle

import arcpy


class LicenseError(Exception):
    """Reprensents a failure when performing license checks.

    Public Attributes:
    product: code for the product that failed license check.
    extensions: list of codes for the extensions that failed license check
    """

    def __init__(self, product=None, extensions=None, message=None):
        """Initialize public class attributes."""
        super(LicenseError, self).__init__(message)
        self.product = product
        self.extensions = extensions
        self.message = message if message else "Required license(s): "

    def __str__(self):
        """Provide a string representation of invalid product and extension codes."""
        failed_codes = []
        if self.product:
            failed_codes.append(self.product)
        if self.extensions:
            if isinstance(self.extensions, (list, tuple)):
                failed_codes += list(self.extensions)
            else:
                failed_codes.append(self.extensions)

        return "{} {}".format(self.message, ", ".join(failed_codes))


class NATool(object):
    """Base class used by all the network analysis sample tools.

    It contains methods that are used by many network analysis sample tools.
    """

    # Error message strings
    MSG_INVALID_NETWORK_SOURCE = "Network source does not support travel modes or is not valid."
    MSG_NOT_SIGNED_IN = "Sign in to ArcGIS Online and set it as your active portal."
    MSG_NOT_SIGNED_AGOL = "Set ArcGIS Online as your active portal."
    MSG_CUSTOM_TRAVEL_MODE = "Using a custom travel mode."

    @staticmethod
    def is_arcgis_pro():
        """Return true if using ArcGIS Pro."""
        install_info = arcpy.GetInstallInfo()
        product_name = install_info["ProductName"]
        return True if product_name.lower() == "arcgispro" else False

    @staticmethod
    def import_service(service_name, username="", password="", ags_connection_file="", token="", referer=""):
        """Import the service toolbox based on the specified credentials and returns the toolbox object.

        Args:
            service_name: Name of the service including the service folder that should be imported. For example, to
                           import a service called ServiceAreas that is available in a service folder called World, the
                           service_name should be 'World/ServiceAreas'. If the service is available in the root folder,
                           service_name should only include the name of the service.
            username: The name of the user that has access to the service if the service is secured. For unsecured
                       service, username should not be passed.
            password: Password for the user that can access the service if the service is secured. For unsecured
                       service, password should not be passed.
            ags_connection_file: Full catalog path to an ArcGIS Server Connection file. The connection file should be
                                 be created with the username and password for the user that has access to the service
                                 if the service is secured.
            token: A string representing the access credentials for the user that has access to the service.
            referer: A string representing the HTTP Referer that was used when obtaining the token.

        Returns:
            A module object that can be used to call all tools within the service.

        Raises:
            If there is an issue when importing the service such as if the user name or password is invalid, the
            function raises an arcpy.ExecuteError
        """
        # Construct the connection string to import the service toolbox
        if username and password:
            tbx = "https://logistics.arcgis.com/arcgis/services;{0};{1};{2}".format(service_name, username, password)
        elif ags_connection_file:
            tbx = "{0};{1}".format(ags_connection_file, service_name)
        elif token and referer:
            tbx = "https://logistics.arcgis.com/arcgis/services;{0};token={1};{2}".format(service_name, token, referer)
        else:
            raise arcpy.ExecuteError("No valid option specified to connect to the {0} service".format(service_name))

        # Import the service toolbox
        try:
            arcpy.ImportToolbox(tbx)
            return getattr(arcpy, service_name.split("/")[-1])
        except Exception as ex:
            raise arcpy.ExecuteError("Failed to import the service toolbox\n{}".format(ex))

    @staticmethod
    def get_travel_modes(network_source="ARCGIS_ONLINE"):
        """Return a list of travel modes supported by the network source and the default travel mode name.

        As retrieving travel modes from ArcGIS Online can be slow, the function stores the travel mode from ArcGIS
        Online in a file and returns the travel modes from the file in case the file exists. So making subsequent calls
        to the function should be relatively faster.

        Args:
            network_source: The street data used to perform the analysis. The value can be a network dataset
                            (specified as the catalog path or a layer object) or a string keyword, 'ARCGIS_ONLINE'
                            indicating the use of ArcGIS Online services.
        Returns:
            A dictionary containing the travel modes supported by the network source and the default travel mode set
            on the network source. The dictionary contains two keys, 'supportedTravelModes' whose value is a dictionary
            of travel mode names and travel mode settings, and 'defaultTravelModeName' whose value is the name of the
            travel mode to be used as default.

        Raises:
            When using ArcGIS Online as the network source, an arcpy.ExecuteError is raised if not signed into a portal.
            A ValueError is raised if the network_source is not ARCGIS_ONLINE or a network dataset that exists.

        """
        travel_modes = {
            "supportedTravelModes": {},
            "defaultTravelModeName": "",
        }
        supported_travel_modes = {}
        default_travel_mode_name = ""
        is_arcgis_online = False
        token = ""
        # Check if using ArcGIS Online as the network source
        if hasattr(network_source, "upper"):
            if network_source.upper() == "ARCGIS_ONLINE":
                is_arcgis_online = True
                access_credentials = arcpy.GetSigninToken()
                if not access_credentials:
                    raise arcpy.ExecuteError("Please sign in to ArcGIS Online")
                token = access_credentials["token"]
                # It is expensive to retrieve travel modes from ArcGIS Online. So check if the travel modes were saved
                # from earlier calls to the function. If a saved version is found return travel modes from the saved
                # version.
                travel_modes_file_name = hashlib.md5(token.encode("utf-8")).hexdigest()
                # Save the travel modes file in the temp folder used by the application so that it will deleted when
                # the application is closed. The application specific temp folder is different in ArcGIS Pro and ArcMap
                if NATool.is_arcgis_pro():
                    app_temp_folder = os.getenv("TEMP")
                else:
                    app_temp_folder = arcpy.env.packageWorkspace  # pylint: disable=no-member
                travel_modes_file = os.path.join(app_temp_folder, travel_modes_file_name)
                if os.path.exists(travel_modes_file):
                    with open(travel_modes_file, "rb") as travel_modes_fp:
                        travel_modes = pickle.load(travel_modes_fp)
                else:
                    # Get the travel modes from the GetTravelModes tool on Utilities service.
                    try:
                        utilities_service = NATool.import_service("World/Utilities", token=token,
                                                                  referer=access_credentials["referer"])
                    except arcpy.ExecuteError:
                        raise
                    travel_modes_result = utilities_service.GetTravelModes()
                    travel_modes_table = travel_modes_result.getOutput(0)
                    default_travel_mode_id = travel_modes_result.getOutput(1)
                    default_travel_mode_name = ""
                    # Get a dictionary of travel mode names and travel mode properties
                    with arcpy.da.SearchCursor(travel_modes_table,  # pylint: disable=no-member
                                               ("Name", "TravelMode", "TravelModeId")) as cursor:
                        for row in cursor:
                            if row[2] == default_travel_mode_id:
                                default_travel_mode_name = row[0]
                            supported_travel_modes[row[0]] = row[1]
                    travel_modes["supportedTravelModes"] = supported_travel_modes
                    travel_modes["defaultTravelModeName"] = default_travel_mode_name
                    # Save the travel modes to a file
                    with open(travel_modes_file, "wb") as travel_modes_fp:
                        pickle.dump(travel_modes, travel_modes_fp)
        # Assume network_source is a local network dataset
        if not is_arcgis_online:
            # Determine the catalog path of the network_source since arcpy.na.GetTravelModes needs a catalog path and
            # network_source can be a network dataset layer. A network dataset layer object has a dataSource property
            nds_catalog_path = network_source
            if hasattr(network_source, "dataSource"):
                nds_catalog_path = network_source.dataSource  # pylint: disable=no-member
            elif hasattr(network_source, "value"):
                nds_catalog_path = network_source.value  # pylint: disable=no-member
            try:
                supported_travel_modes = arcpy.na.GetTravelModes(nds_catalog_path)
            except ValueError:
                raise
            # Determine the default travel mode name for the network dataset. If the network dataset does not have a
            # default travel mode, select the first element from the list of supported travel modes.
            nds_desc = arcpy.Describe(network_source)
            default_travel_mode_name = nds_desc.defaultTravelModeName  # pylint: disable=no-member
            if not default_travel_mode_name:
                if supported_travel_modes:
                    default_travel_mode_name = list(supported_travel_modes)[0]

            travel_modes["supportedTravelModes"] = supported_travel_modes
            travel_modes["defaultTravelModeName"] = default_travel_mode_name

        return travel_modes

    @staticmethod
    def verify_network_source(network_source_param, travel_mode_param):
        """Check if signed in to ArcGIS Online or a local network source is valid and supports travel modes.

        The function should only be called from the updateMessages method in the ToolValidation class for the tool.

        Args:
            network_source_param: The object referencing the network source tool parameter.
            travel_mode_param: The object referencing the travel mode tool parameter.
        """
        network_source = network_source_param.value
        travel_mode_param_filter = travel_mode_param.filter.list

        if network_source:
            if network_source == "ARCGIS_ONLINE":
                if arcpy.GetSigninToken():
                    # Raise an error if signed into some other portal and arcgis.com is not your active portal
                    active_portal_url = arcpy.GetActivePortalURL().rstrip("/")
                    if active_portal_url.lower() not in ("http://www.arcgis.com", "https://www.arcgis.com"):
                        network_source_param.setErrorMessage(NATool.MSG_NOT_SIGNED_AGOL)
                else:
                    # Raise an error if not signed in to ArcGIS Online
                    network_source_param.setErrorMessage(NATool.MSG_NOT_SIGNED_IN)
            else:
                # Raise an error if the local network source does not support travel modes
                if len(travel_mode_param_filter) == 0:
                    network_source_param.setErrorMessage(NATool.MSG_INVALID_NETWORK_SOURCE)

    @staticmethod
    def verify_travel_mode(travel_mode_param):
        """Check if the travel mode value is a valid travel mode JSON or one of the supported travel mode names.

        The function should only be called from the updateMessages method in the ToolValidation class for the tool.

        Args:
            travel_mode_param: The object referencing the travel mode tool parameter.
        """
        travel_mode = travel_mode_param.value
        travel_mode_param_filter = travel_mode_param.filter.list
        # ERROR 000800: Value is not a member of
        if travel_mode_param.hasError() and travel_mode_param.message.find("000800") > -1:
            try:
                arcpy.na.TravelMode(travel_mode)
                travel_mode_param.setWarningMessage(NATool.MSG_CUSTOM_TRAVEL_MODE)
            except ValueError:
                travel_mode_param.setIDMessage("ERROR", 30158, travel_mode, u" | ".join(travel_mode_param_filter))

    @staticmethod
    def check_license(product="arcview", extensions=None):
        """Return true if the required product and extension licenses are initialized.

        If either product or license checks are unsuccessful, LicenseError is raised. The product and extensions
        properties of the Exception can be used to determine the required licenses that were not code.

        Args:
            product: code for the minimum product level required for the tool. Valid values are same as those supported
                     by arcpy.CheckProduct http://pro.arcgis.com/en/pro-app/arcpy/functions/checkproduct.htm
            extensions: list of extension codes required to run the the tool. valid values are same as those supported
                    by arcpy.CheckExtension http://pro.arcgis.com/en/pro-app/arcpy/functions/checkextension.htm

        Returns:
            True if both product and extension license checks are successfull.

        """
        product_check = False
        failed_product = ""
        extension_check = False
        failed_extensions = []

        # Check if the product has been already initialized. AlreadyInitialized is returned if a lower product code is
        # checked but the product is using a higher level license. For example, if running as Advanced and checking for
        # Basic.
        if arcpy.CheckProduct(product) in ("AlreadyInitialized", "Available"):
            product_check = True
        else:
            product_check = False
            failed_product = product

        # Check if the extension licenses are available
        if extensions:
            if not isinstance(extensions, (list, tuple)):
                extensions = extensions.split(",")
            failed_extensions = [extension for extension in extensions
                                 if arcpy.CheckExtension(extension) != "Available"]
            extension_check = False if failed_extensions else True

        if all((product_check, extension_check)):
            return True
        else:
            raise LicenseError(failed_product, failed_extensions)